
/*
                                +---------------------------------+
                                |                                 |
                                |         ***   BSP   ***         |
                                |                                 |
                                |  Copyright  -tHE SWINe-  2002  |
                                |                                 |
                                |             Bsp.cpp             |
                                |                                 |
                                +---------------------------------+
*/

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Load_3ds.h"
#include "Matrix.h"
#include "Main.h"
#include "List.h"
#include "Bsp.h"

#define Epsilon (float)0.001

#define _Len(A) ((float)sqrt(A.x*A.x+A.y*A.y+A.z*A.z))
// dlka vektoru

#define _Normalize(A) {float s=LEN(A); A.x/=s;A.y/=s;A.z/=s;}
// vytvo jednotkov vektor

#define _Cross(A,B,C) C.x = B.y*A.z-A.y*B.z; \
                      C.y = B.z*A.x-A.z*B.x; \
                      C.z = B.x*A.y-A.x*B.y;
// skalrn souin

static int level;
static int mx_level;
static int poly_cnt;
static int poly_done;
static int nod_num;

// vytvori bsp-strom z 3ds-sveta
BSP_Tree *CreateTree(World *svet)
{
    BSP_Tree *tmp;
    long start;
    Face *face;
    int i, j;

    start = clock();

    printf("   Creating BSP_Tree\n");
    if((tmp = (BSP_Tree*)malloc(sizeof(BSP_Tree))) == NULL) {
        ErrMsg("Not enough memory to create a tree structure !");

        return NULL;
    }
    // naalokuje pam pro bsp-strom

    tmp->n_material_num = svet->n_material_num;
    if((tmp->material = (BSP_Material*) malloc
       (tmp->n_material_num * sizeof(BSP_Material))) == NULL) {
        ErrMsg("Not enough memory to create a tree structure !");
        return NULL;
    }
    for(i = 0; i < tmp->n_material_num; i ++)
        strcpy(tmp->material[i].name, svet->material[i].name);
    // zkopruje jmna materil (textur) ;-)

    Init_VertexList(&tmp->vertex);
    printf("     Creating Vertex - list  \\");
    for(i = 0; i < svet->n_object_num; i ++) {
        for(j = 0; j < svet->object[i].n_vertex_num; j ++)
            AddVertexToList(&svet->object[i].vertex[j], &tmp->vertex);
        UpdateStatus();
    }
    printf("\n");
    // vytvo vertex - list

    Init_FaceList(&tmp->face);
    printf("     Adding faces to list  \\");
    for(i = 0; i < svet->n_object_num; i ++) {
        for(j = 0; j < svet->object[i].n_face_num; j ++) {
            face = AddFaceToList(&svet->object[i].face[j], &tmp->face);
            face->vertex[0] = tmp->vertex.vertex[FindVertexInList
               (svet->object[i].face[j].vertex[0], &tmp->vertex)];
            face->vertex[1] = tmp->vertex.vertex[FindVertexInList
               (svet->object[i].face[j].vertex[1], &tmp->vertex)];
            face->vertex[2] = tmp->vertex.vertex[FindVertexInList
               (svet->object[i].face[j].vertex[2], &tmp->vertex)];
            UpdateStatus();
        }
    }
    printf("\r     Adding faces to list\n");
    // vytvo face - list

    for(i = 0; i < tmp->face.n_face_num; i ++)
        tmp->face.face[i]->n_id = i;
    // piad kadmu face id

    if((tmp->root = (BSP_Node*) malloc(sizeof(BSP_Node))) == NULL) {
        ErrMsg("Not enough memory to save a root node structure !");
        return NULL;
    }
    memset(tmp->root, 0, sizeof(BSP_Node));
    tmp->root->vertex = &tmp->vertex;
    memcpy(&tmp->root->face, &tmp->face, sizeof(Face_List));
    // vytvo root - node (koen) stromu
    
    level = 0;
    mx_level = 0;
    poly_cnt = tmp->face.n_face_num;
    poly_done = 0;
    nod_num = 0;
    CreateNode(tmp->root);
    // vytv rekurzivn dal subnody

    printf("\n     Done [%d secs, %d nodes, %d faces, %d width]\n",
       (clock() - start) / CLOCKS_PER_SEC, nod_num, poly_cnt, mx_level);
    // vytiskne jak dlouho to trvalo, poet fac ..

    return tmp;
}

// vykresluje ukazatel pracovani
// - toc caru
void UpdateStatus()
{
    char stat[] = "/|\\-";
    static int a = 0, b = 0;
    
    if(b ++ > 50) {
        b = 0;
        a ++;
        if(a > 3)
            a = 0;
        printf("\b%c", stat[a]);
    }
    // kresl "mrnc" se ru
}

// spot nodu bsp-stromu
// pi tom tiskne informace
// o stavu operace, potu poly ...
// nod je noda kter se m spotat.
int CreateNode(BSP_Node *nod)
{
    nod_num ++;
    level ++;
    if(level > mx_level)
        mx_level = level;
    // pot rove a vku stromu

    printf("\r     Level %3d -ALC- %5d Faces [%3d%%]  ",
        level, nod->face.n_face_num, (int)(0.5 +
        (float)poly_done / (float)poly_cnt * 100));
    //
    if((nod->front = (BSP_Node*)malloc(sizeof(BSP_Node))) == NULL)
        return 0;
    //
	memset(nod->front, 0, sizeof(BSP_Node));
    memcpy(&nod->front->face, &nod->face, sizeof(Face_List));
    nod->front->vertex = nod->vertex;
    //
    if((nod->back = (BSP_Node*)malloc(sizeof(BSP_Node))) == NULL)
        return 0;
    //
	memset(nod->back, 0, sizeof(BSP_Node));
    memcpy(&nod->back->face, &nod->face, sizeof(Face_List));
    nod->back->vertex = nod->vertex;
    // alokuje pam pro nodu, subnody

    printf("\r     Level %3d -FRP- %5d Faces [%3d%%]  ",
        level, nod->face.n_face_num, (int)(0.5 +
        (float)poly_done / (float)poly_cnt * 100));
    //
    nod->rootpoly = FindRootPoly(nod);
    //
    if(nod->rootpoly == -1) { // noda jue konvexn
        nod->flags = 0;
        nod->onplane = nod->face;
        poly_done += nod->onplane.n_face_num;
        //
        printf("\r     Level %3d -BCK- %5d Faces [%3d%%]  ",
            level, nod->face.n_face_num, (int)(0.5 +
            (float)poly_done / (float)poly_cnt * 100));
        //
        level --;

        return 1;
    }
    memcpy(&nod->rootpl, &nod->face.face[nod->rootpoly]->normal, sizeof(Plane));
    // hled root face stromu

    printf("\r     Level %3d -PPL- %5d Faces [%3d%%]  ",
        level, nod->face.n_face_num, (int)(0.5 +
        (float)poly_done / (float)poly_cnt * 100));
    PartitionSpace(nod);
    poly_done += nod->onplane.n_face_num;
    // rozdluje prostor podle roviny root

    if(nod->flags & _Have_Front_Node) {
        if(!CreateNode(nod->front)) {
            printf("\r     Level %3d -BCK- %5d Faces [%3d%%]  ",
                level, nod->face.n_face_num, (int)(0.5 +
                (float)poly_done / (float)poly_cnt * 100));
            level --;

            return 0;
        }
    }
    // vytvo pedn nodu

    if(nod->flags & _Have_Back_Node) {
        if(!CreateNode(nod->back)) {
            printf("\r     Level %3d -BCK- %5d Faces [%3d%%]  ",
                level, nod->face.n_face_num, (int)(0.5 +
                (float)poly_done / (float)poly_cnt * 100));
            level --;

            return 0;
        }
    }
    // vytvo zadn nodu

    printf("\r     Level %3d -BCK- %5d Faces [%3d%%]  ",
        level, nod->face.n_face_num, (int)(0.5 +
        (float)poly_done / (float)poly_cnt * 100));
    level --;
    // jde o patro n

    return 1;
}

// vrt index root polygonu nody nod,
// nebo -1 pokud je noda konvexn (=> nen
// ji u poteba dl dlit)
int FindRootPoly(BSP_Node *nod)
{
    int n_tot_f_num, n_tot_b_num, n_tot_s_num;
    int n_min_score, n_cur_score, n_min_i;
    int n_f_num, n_b_num, n_s_num;
    int i, j;

    n_min_i = -1;
    n_min_score = 0xffff;
    // nastav promnn pro hledn nejlepho skre

    n_tot_f_num = 0;
    n_tot_b_num = 0;
    n_tot_s_num = 0;
    // pomoc tchhle promnnch se zjist, zda je noda konvexn

    for(i = 0; i < nod->face.n_face_num; i ++) {
        n_f_num = 0;
        n_b_num = 0;
        n_s_num = 0;
        // vynuluje potadla

        for(j = 0; j < nod->face.n_face_num; j ++) {
            if(i == j)
                continue;
            switch(ClassifyFacePl(&nod->face.face[i]->normal,
               nod->face.face[j])) {
            case _Front:
                n_f_num ++;
                break;
            case _Back:
                n_b_num ++;
                break;
            case _Split:
                n_s_num ++;
            default:
                break;
            }
        }
        // spot poty fac za, ped a okolo roviny

        n_tot_f_num += n_f_num;
        n_tot_s_num += n_s_num;
        n_tot_b_num += n_b_num;
        // pite poty fac k totlnmu soutu

        n_cur_score = abs(n_f_num - n_b_num) * 5 + n_s_num * 75;
        if(n_cur_score < n_min_score) {
            n_min_score = n_cur_score;
            n_min_i = i;
        }
        // porovna skre s minimlnm

        UpdateStatus();
    }

    if(n_tot_s_num == 0 && (n_tot_f_num == 0 || n_tot_b_num == 0))
        return -1;
    // vtev je konvexn (tvo kuloid / krychloid / ..)

    return n_min_i;
}


// rozdl polygony v nod
// podle jeji root plane
// do facelist jejich dt
int PartitionSpace(BSP_Node *nod)
{
    int i;

    Init_FaceList(&nod->front->face);
    Init_FaceList(&nod->back->face);
    Init_FaceList(&nod->onplane);
    // nastav facelisty na prdn

    nod->flags = 0;
    // nastav nodu na "bezdtnou"

    for(i = 0; i < nod->face.n_face_num; i ++) {
        // vezme face a podiv se, kde lezi
        switch(ClassifyFacePl(&nod->rootpl, nod->face.face[i])) {
        case _OnPlane:
            AddFaceToList(nod->face.face[i], &nod->onplane);
            // le v rovine s root face (=> zstane
            // v rodiovsk nod jako "OnPlane")
            break;
        case _Front:
            AddFaceToList(nod->face.face[i], &nod->front->face);
            nod->flags |= _Have_Front_Node;
            // le ped rovinou root facu, zrove
            // nastav pedn vtev na plnou
            break;
        case _Back:
            AddFaceToList(nod->face.face[i], &nod->back->face);
            nod->flags |= _Have_Back_Node;
            // le za rovinou root facu, zrove
            // nastav zadn vtev na plnou
            break;
        case _Split:
            // ppadn face podle roviny rozdl na dal (face, ne polygony !) ...
            if(!SplitPolygon(nod, nod->face.face[i],
               &nod->front->face, &nod->back->face)) {
                ErrMsg("Internal program error (maybe windows found ?)");
                return 0;
            }
            nod->flags |= _Have_Front_Node;
            nod->flags |= _Have_Back_Node;
            // a ozna ob vtve za pln ...
            break;
        default:
            return 0;
            break;
        }
        UpdateStatus();
        // vytiskne ukazatel akce
    }

    return 1;
}

// vypot prsek roviny a seky,
// definovan body p_v1 a p_v2
Vertex IntEdgePlane(Vertex *p_v1, Vertex *p_v2, Plane *p_plane)
{
    float f_a, f_b, f_t;
    Vertex vertex;

    f_a = p_plane->a * p_v1->x + p_plane->b * p_v1->y +
          p_plane->c * p_v1->z + p_plane->d;
    f_b = p_plane->a * p_v2->x + p_plane->b * p_v2->y +
          p_plane->c * p_v2->z + p_plane->d;
    // spot vzdlenosti vertex od roviny

    if(f_a == f_b)
        return *p_v1;
    // singulrn ppad (oba le ve stejn vzdlenosti
    // od roviny na stejn stran, nebo na rovin)

    f_t = f_a / (f_a - f_b);
    // pomr vzdlenost

    vertex.x = p_v1->x + f_t * (p_v2->x - p_v1->x);
    vertex.y = p_v1->y + f_t * (p_v2->y - p_v1->y);
    vertex.z = p_v1->z + f_t * (p_v2->z - p_v1->z);
    vertex.u = p_v1->u + f_t * (p_v2->u - p_v1->u);
    vertex.v = p_v1->v + f_t * (p_v2->v - p_v1->v);
    // musi se spotat i odpovdajc souadnice textur

    return vertex;
}

// rozdl polygon rovinou nod->rootpl
// a vsledky vrt do front a back facelistu
int SplitPolygon(BSP_Node *p_node, Face *p_face, Face_List *list_front, Face_List *list_back)
{
    int i, n_f_num, n_b_num, n_o_num, n_i_front, n_i_back;
    Vertex *v_one, *v_two, v_intersection;
    Face tmp;

    n_f_num = 0;
    n_b_num = 0;
    n_o_num = 0;
    for(i = 0; i < 3; i ++) {
        switch(ClassifyVertexPl(&p_node->rootpl, p_face->vertex[i])) {
        case _Front:
            n_i_front = i;
            n_f_num ++;
            break;
        case _Back:
            n_i_back = i;
            n_b_num ++;
            break;
        case _OnPlane:
            n_i_front = i;
            n_f_num ++;
            n_o_num ++;
            break;
        }
    }
    // spot poty vertex v zvislosti
    // na poloze a zapamatuje si vertex
    // lec na, ped a za rovinou

    tmp.n_id = p_face->n_id;
    tmp.n_material = p_face->n_material;
    tmp.normal = p_face->normal;
    // nastav spolen vlastnosti fac

    if(n_f_num == 3) {
        AddFaceToList(p_face, list_front);
        poly_cnt ++;
        // vechno je ped rovinou / na n

        return 1;
    } else if(n_b_num == 3 || ((n_f_num - n_o_num) == 0)){
        AddFaceToList(p_face, list_back);
        poly_cnt ++;
        // vechno je za rovinou / na n

        return 1;
    } else if(n_f_num == 1 && n_o_num == 0) {
        v_intersection = IntEdgePlane(p_face->vertex[n_i_front],
           p_face->vertex[(n_i_front + 1) % 3], &p_node->rootpl);
        v_one = AddVertexToList(&v_intersection, p_node->vertex);
        //
        v_intersection = IntEdgePlane(p_face->vertex[(n_i_front + 2) % 3],
           p_face->vertex[n_i_front], &p_node->rootpl);
        v_two = AddVertexToList(&v_intersection, p_node->vertex);
        // spot prseky a pid je do globlnho seznamu vrchol

        poly_cnt += 2;
        // zv poet polygon o dva

        tmp.vertex[0] = p_face->vertex[n_i_front];
        tmp.vertex[1] = v_one;
        tmp.vertex[2] = v_two;
        AddFaceToList(&tmp, list_front);
        //
        tmp.vertex[0] = v_one;
        tmp.vertex[1] = p_face->vertex[(n_i_front + 1) % 3];
        tmp.vertex[2] = p_face->vertex[(n_i_front + 2) % 3];
        AddFaceToList(&tmp, list_back);
        //
        tmp.vertex[0] = p_face->vertex[(n_i_front + 2) % 3];
        tmp.vertex[1] = v_two;
        tmp.vertex[2] = v_one;
        AddFaceToList(&tmp, list_back);
        // jeden vertex je ped rovinou, ostatn jsou za

        return 1;
    } else if(n_b_num == 1 && n_o_num != 2) {
        v_intersection = IntEdgePlane(p_face->vertex[n_i_back],
           p_face->vertex[(n_i_back + 1) % 3], &p_node->rootpl);
        v_one = AddVertexToList(&v_intersection, p_node->vertex);
        //
        v_intersection = IntEdgePlane(p_face->vertex[(n_i_back + 2) % 3],
           p_face->vertex[n_i_back], &p_node->rootpl);
        v_two = AddVertexToList(&v_intersection, p_node->vertex);
        // spot prseky a pid je do globlnho seznamu vrchol

        poly_cnt += 2;
        // zven potu polygon o dva

        tmp.vertex[0] = v_one;
        tmp.vertex[1] = p_face->vertex[(n_i_back + 1) % 3];
        tmp.vertex[2] = v_two;
        AddFaceToList(&tmp, list_front);
        //
        tmp.vertex[0] = p_face->vertex[(n_i_back + 1) % 3];
        tmp.vertex[1] = p_face->vertex[(n_i_back + 2) % 3];
        tmp.vertex[2] = v_two;
        AddFaceToList(&tmp, list_front);
        //
        tmp.vertex[0] = v_one;
        tmp.vertex[1] = v_two;
        tmp.vertex[2] = p_face->vertex[n_i_back];
        AddFaceToList(&tmp, list_back);
        // jeden vertex je za rovinou, ostatn ped

        return 1;
    }

    return 0;
}

// klasifikuje vertex proti rovin
// vsledky _Front - vpedu, _Back - vzadu
// a _OnPlane - na rovin
int ClassifyVertexPl(Plane *p_plane, Vertex *p_vertex)
{
    float p_eq;

    p_eq = p_plane->a * p_vertex->x + p_plane->b * p_vertex->y +
           p_plane->c * p_vertex->z + p_plane->d;
    // spot rovnici roviny

    if(p_eq > Epsilon)
        return _Front;
    if(p_eq < -Epsilon)
        return _Back;

    return _OnPlane;
}

// klasifikuje face proti rovin
// vsledky _Front - vpedu, _Back - vzadu
// a _OnPlane - na rovin a _Split - protn rovinu
int ClassifyFacePl(Plane *p_plane, Face *p_face)
{
    int i, n_f_num = 0, n_b_num = 0, n_o_num = 0;

    for(i = 0; i < 3; i ++) {
        switch (ClassifyVertexPl(p_plane, p_face->vertex[i])) {
        case _Front:
            n_f_num ++;
            break;
        case _Back:
            n_b_num ++;
            break;
        default: // _OnPlane
            n_o_num ++;
            break;
        }
    }
    // spot poty vertex ped, za a na rovin

    // vyhodnot vsledek
    if(n_o_num == 3)
        return _OnPlane;
    if(n_f_num == 3)
        return _Front;
    if(n_b_num == 3)
        return _Back;
    if(n_f_num + n_o_num == 3)
        return _Front;
    if(n_b_num + n_o_num == 3)
        return _Back;

    return _Split;
}

// zape obsah nody nod do souboru fw
void WriteNode(BSP_Node *nod, FILE *BSP_fp)
{
    __int16 int16[1];
    int i;

    // zapise nodu do souboru
    level ++;
    printf("\r     Writing node -F- Level %3d", level);

    int16[0] = nod->flags;
    fwrite(int16, sizeof(int16), 1, BSP_fp);
    // zapise vlastnosti nody

    fwrite(&nod->rootpl.a, sizeof(float), 1, BSP_fp);
    fwrite(&nod->rootpl.b, sizeof(float), 1, BSP_fp);
    fwrite(&nod->rootpl.c, sizeof(float), 1, BSP_fp);
    fwrite(&nod->rootpl.d, sizeof(float), 1, BSP_fp);
    // zapise normalu root facu

    int16[0] = nod->onplane.n_face_num;
    fwrite(int16, sizeof(int16), 1, BSP_fp);
    // zapise pocet facu

    for(i = 0; i < nod->onplane.n_face_num; i ++) {
        int16[0] = FindVertexInList(nod->onplane.face[i]->vertex[0], nod->vertex);
        fwrite(int16, sizeof(int16), 1, BSP_fp);
        //
        int16[0] = FindVertexInList(nod->onplane.face[i]->vertex[1], nod->vertex);
        fwrite(int16, sizeof(int16), 1, BSP_fp);
        //
        int16[0] = FindVertexInList(nod->onplane.face[i]->vertex[2], nod->vertex);
        fwrite(int16, sizeof(int16), 1, BSP_fp);
        // veretx_id

        int16[0] = nod->onplane.face[i]->n_id;
        fwrite(int16, sizeof(int16), 1, BSP_fp);
        // id

        int16[0] = nod->onplane.face[i]->n_material;
        fwrite(int16, sizeof(int16), 1, BSP_fp);
        // material

        fwrite(&nod->onplane.face[i]->normal.a, sizeof(float), 1, BSP_fp);
        fwrite(&nod->onplane.face[i]->normal.b, sizeof(float), 1, BSP_fp);
        fwrite(&nod->onplane.face[i]->normal.c, sizeof(float), 1, BSP_fp);
        fwrite(&nod->onplane.face[i]->normal.d, sizeof(float), 1, BSP_fp);
        // normala
    }
    // zapisuje facy

    if(nod->flags & _Have_Front_Node)
        WriteNode(nod->front, BSP_fp);

    if(nod->flags & _Have_Back_Node)
        WriteNode(nod->back, BSP_fp);
    // zapisuje dalsi subnody

    printf("\r     Writing node -B- Level %3d", level);
    level --;
}

// zape strom tree do souboru na adrese f_name
// pouv se binrn zpis a pi ten je to nutn
// dodret (zape se pesn to, co m, k CR se
// nedopluje LF a pi ten se ignoruje EOF)
int WriteTree(BSP_Tree *tree, char *f_name)
{
    __int16 int16[1];
    char chr[3];
    FILE *BSP_fp;
    int i;

    printf("   Writing BSP_Tree\n");
    printf("     Creating output file\n");
    if((BSP_fp = fopen(f_name, "wb")) == NULL) {
        ErrMsg("Cannot create file - disk full / locked ?");
        return 0;
    }
    // oteve soubor pro zpis

    chr[0] = 'B';
    chr[1] = 'S';
    chr[2] = 'P';
    printf("     Writing header\n");
    fwrite(chr, sizeof(chr), 1, BSP_fp);
    // zape typ souboru - "BSP"

    int16[0] = 50;
    fwrite(int16, sizeof(int16), 1, BSP_fp);
    // zape verzi souboru - 50

    int16[0] = tree->vertex.n_vertex_num;
    fwrite(int16, sizeof(int16), 1, BSP_fp);
    // poet vertex

    int16[0] = tree->face.n_face_num;
    fwrite(int16, sizeof(int16), 1, BSP_fp);
    // zape poet fac

    int16[0] = tree->n_material_num;
    fwrite(int16, sizeof(int16), 1, BSP_fp);
    // poet materil

    for(i = 0; i < tree->n_material_num; i ++)
        fwrite(tree->material[i].name, 256, 1, BSP_fp);
    // zape jmena materil

    for(i = 0; i < tree->vertex.n_vertex_num; i ++) {
        fwrite(&tree->vertex.vertex[i]->x, sizeof(float), 1, BSP_fp);
        fwrite(&tree->vertex.vertex[i]->y, sizeof(float), 1, BSP_fp);
        fwrite(&tree->vertex.vertex[i]->z, sizeof(float), 1, BSP_fp);
        fwrite(&tree->vertex.vertex[i]->u, sizeof(float), 1, BSP_fp);
        fwrite(&tree->vertex.vertex[i]->v, sizeof(float), 1, BSP_fp);
        // souadnice

        fwrite(&tree->vertex.vertex[i]->normal.x, sizeof(float), 1, BSP_fp);
        fwrite(&tree->vertex.vertex[i]->normal.y, sizeof(float), 1, BSP_fp);
        fwrite(&tree->vertex.vertex[i]->normal.z, sizeof(float), 1, BSP_fp);
        // normly vertex - dleit ! (kdy se svt
        // tolikrt perozdl, je tk normly sprvn zkonstruovat)
    }
    // zapise pole vech vertex

    level = 0;
    WriteNode(tree->root, BSP_fp);
    // zape rekurzivn vechny nody

    printf("\n     Done\n");
    fclose(BSP_fp);
    // uzave soubor

    return 1;
}
